
/****************************************************************************
 *
 * MODULE:       v.mk_circle
 * AUTHOR(S):    Bob Covill <tekmap.ns.ca>
 * PURPOSE:      Create a vector file of circles at user defined locations and size
 * COPYRIGHT:    (C) 1999-2012 by the GRASS Development Team
 *
 *               This program is free software under the GNU General Public
 *               License (>=v2). Read the file COPYING that comes with GRASS
 *               for details.
 *
 *****************************************************************************/

#include <stdlib.h>
#include <strings.h>
#include <math.h>
#include <grass/gis.h>
#include <grass/dbmi.h>
#include <grass/vector.h>
#include <grass/glocale.h>

#define BUF_SIZE 2000

void save_vect_feature(struct Map_info *, double, double, double, double,
		       int);

double get_real_radius(double, double, double, double);

int main(int argc, char *argv[])
{
    struct GModule *module;

    struct Option *coord, *delim, *ifile, *dist, *new, *shape;

    struct Flag *centroid;

    double east, north, radius;

    double angle_increment = 2.0;	/* determines accuracy of circle and fibnal shape 
					   120.0 = triangle, 90 = diamond, 72 = pentagon, 60 = hexagon, 45.0 = octagon */

    struct Map_info Map;

    char *fs;

    char *buf, *buf_raw;

    char *coorbuf;

    int ntokens;

    char **tokens;

    FILE *in_fd;

    G_gisinit(argv[0]);

    module = G_define_module();
    G_add_keyword("vector");
    module->description =
	_("Generate closed vector from center coordinate and size.");

    new = G_define_standard_option(G_OPT_V_OUTPUT);

    coord = G_define_option();
    coord->key = "coordinate";
    coord->type = TYPE_STRING;
    coord->required = NO;
    coord->key_desc = "x,y";
    coord->description = _("The center coordinate (east,north)");

    dist = G_define_option();
    dist->key = "radius";
    dist->type = TYPE_DOUBLE;
    dist->required = NO;
    dist->description = _("The radius of the output shape.");

    ifile = G_define_standard_option(G_OPT_F_INPUT);
    ifile->required = NO;
    ifile->label = _("Name of input file to be imported");
    ifile->description = ("'-' for standard input");


    delim = G_define_standard_option(G_OPT_F_SEP);
    delim->required = NO;
    delim->guisection = _("Input format");


    shape = G_define_option();
    shape->key = "shape";
    shape->type = TYPE_STRING;
    shape->required = NO;
    shape->multiple = NO;
    shape->options = "circle,octagon,hexagon,pentagon,diamond,triangle";
    shape->answer = "circle";
    shape->description = _("The shape of ouput vector");

    centroid = G_define_flag();
    centroid->key = 'c';
    centroid->description =
	_("Build shapes as closed centroids (default=open)");

    if (G_parser(argc, argv))
	exit(EXIT_FAILURE);

    if (ifile->answer && coord->answer)
	G_fatal_error(_("Cannot use input file and coordinate arguments together"));

    if (coord->answer && !dist->answer)
	G_fatal_error(_("Must provide a radius argument with the corrdinate argument"));

    if (dist->answer)
	radius = atof(dist->answer);

    angle_increment = 2.0;	/* circle */
    if (strcmp(shape->answer, "octagon") == 0)
	angle_increment = 45.0;
    else if (strcmp(shape->answer, "hexagon") == 0)
	angle_increment = 60.0;
    else if (strcmp(shape->answer, "pentagon") == 0)
	angle_increment = 72.0;
    else if (strcmp(shape->answer, "diamond") == 0)
	angle_increment = 90.0;
    else if (strcmp(shape->answer, "triangle") == 0)
	angle_increment = 120.0;
    else
	angle_increment = 2.0;

    /* Open Vector File for writing */
    Vect_open_new(&Map, new->answer, WITHOUT_Z);	/* not 3D */
    Vect_hist_command(&Map);

    /* Process using single point from command line */
    if (coord->answer) {
	G_scan_easting(coord->answers[0], &east, G_projection());
	G_scan_northing(coord->answers[1], &north, G_projection());

	save_vect_feature(&Map, east, north, radius, angle_increment,
			  centroid->answer);

    }
    else {
	/* Get pints from file or stdin */

	/* Field separator */
	fs = delim->answer;
	if (strcmp(fs, "\\t") == 0)
	    fs = "\t";
	if (strcmp(fs, "tab") == 0)
	    fs = "\t";
	if (strcmp(fs, "space") == 0)
	    fs = " ";
	if (strcmp(fs, "comma") == 0)
	    fs = ",";

	/* Open file */
	if (strcmp(ifile->answer, "-")) {
	    if ((in_fd = fopen(ifile->answer, "r")) == NULL) {
		G_fatal_error(_("Unable to open input file <%s>"),
			      ifile->answer);
	    }
	}
	else {
	    in_fd = stdin;
	}


	/* alloc read buffers */
	buf = (char *)G_malloc(BUF_SIZE);
	buf_raw = (char *)G_malloc(BUF_SIZE);
	coorbuf = (char *)G_malloc(256);

	/* parse file */
	while (1) {
	    if (G_getl2(buf, BUF_SIZE - 1, in_fd) == 0)
		break;		/* EOF */

	    /* Skip comment lines */
	    if ((buf[0] == '#') || (buf[0] == '\0')) {
		continue;
	    }

	    strcpy(buf_raw, buf);

	    tokens = G_tokenize(buf, fs);
	    ntokens = G_number_of_tokens(tokens);

	    if (ntokens < 2)
		G_fatal_error(_("Expected at least 2 columns but got <%d> columns.\nCheck input file or field seperator."),
			      ntokens);

	    sprintf(coorbuf, "%s", tokens[0]);
	    G_scan_easting(coorbuf, &east, G_projection());

	    sprintf(coorbuf, "%s", tokens[1]);
	    G_scan_northing(coorbuf, &north, G_projection());

	    if (ntokens > 2)
	    {
		sscanf(tokens[2], "%lf", &radius);
    	    }
	    else
	    {
		if (!dist->answer)
		    G_fatal_error(_("Cannot determine the radius for the output shape.\nThe radius must be provided in the 3rd column of the input file or given as the radius argument."));
	    }

	    save_vect_feature(&Map, east, north, radius, angle_increment,
			      centroid->answer);

	    G_free_tokens(tokens);
	}

	G_free(buf);
	G_free(buf_raw);
	G_free(coorbuf);
	fclose(in_fd);
    }

    /* Wrap it up */
    Vect_build(&Map);

    /* Vect_set_release_support(&Map); */

    Vect_close(&Map);

    return (0);

}

/**********************************************************
 * Write out shape to vector file using location and radius
***********************************************************/
void
save_vect_feature(struct Map_info *Map,
		  double east,
		  double north,
		  double radius, double angle_increment, int build_centroid)
{
    double angle;

    double ll_val = radius;

    struct line_pnts *Points;

    struct line_cats *Cats;

    Points = Vect_new_line_struct();
    Cats = Vect_new_cats_struct();

    for (angle = 0.0; angle <= 360.0; angle += angle_increment) {
	ll_val = get_real_radius(east, north, radius, angle);
	Vect_append_point(Points,
			  (double)(east + (sin(angle * 0.01745) * ll_val)),
			  (double)(north + (cos(angle * 0.01745) * ll_val)),
			  0.0);
    }
    /* Close circle */
    angle = 0.0;
    ll_val = get_real_radius(east, north, radius, angle);
    Vect_append_point(Points,
		      (double)(east + (sin(angle * 0.01745) * ll_val)),
		      (double)(north + (cos(angle * 0.01745) * ll_val)), 0.0);

    /* Create circle */
    if (build_centroid)
	Vect_write_line(Map, GV_BOUNDARY, Points, Cats);
    else
	Vect_write_line(Map, GV_LINE, Points, Cats);


    /* Add centroid for area */
    if (build_centroid) {
	Vect_reset_line(Points);
	Vect_append_point(Points, east, north, 0.0);
	Vect_cat_set(Cats, 1, 1);
	Vect_write_line(Map, GV_CENTROID, Points, Cats);
    }

}

/**********************************************************
 * Simple function to get convert radius to decimal degrees
 * if required -- otherwise stays the same
***********************************************************/
double get_real_radius(double east, double north, double radius, double azi)
{
    double radius_out = radius;

    if (G_projection() == PROJECTION_LL) {
	/* need to convert radius to degrees */
	double degree_size = 0.5;

	double e2, n2, dist;

	G_begin_distance_calculations();

	e2 = east + (sin(azi * 0.01745) * degree_size);
	n2 = north + (cos(azi * 0.01745) * degree_size);
	dist = G_distance(east, north, e2, n2);
	radius_out = (degree_size * radius) / dist;
    }

    return radius_out;

}
